/* GIOS - General IO set
 * Copyright (C) 2002  Frank Rehberger, Frank.Rehberger@web.de
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#ifndef WIN32

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/poll.h>
#include <sys/ioctl.h>
#include <sys/wait.h>

#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <error.h>
#include <signal.h>
#include <assert.h>
#include <stdarg.h>

#include "gios.h"

/* stolen from gnu-libc */
# define TEMP_FAILURE_RETRY(expression) \
  (__extension__                                                              \
    ({ long int __result;                                                     \
       do __result = (long int) (expression);                                 \
       while (__result == -1L && errno == EINTR);                             \
       __result; }))

gint 
g_ios_init (GError** err)
{
   return 0;
}

static 
void
file_set_error(GError** err, 
	       int      errnum)
{
   gint code = 0;
   if (err==NULL) return;
   /* g_warning ("caught %d\n", errnum); */
   switch (errnum)
   {
      case EEXIST:       code=G_FILE_ERROR_EXIST;  break;
      case EISDIR:       code=G_FILE_ERROR_ISDIR;  break;
      case EACCES:       code=G_FILE_ERROR_ACCES;  break;
      case ENAMETOOLONG: code=G_FILE_ERROR_NAMETOOLONG; break;
      case ENOENT:       code=G_FILE_ERROR_NOENT;  break;
      case ENOTDIR:      code=G_FILE_ERROR_NOTDIR; break;
      case ENXIO:        code=G_FILE_ERROR_NXIO;   break;
      case ENODEV:       code=G_FILE_ERROR_NODEV;  break;
      case EROFS:        code=G_FILE_ERROR_ROFS;   break;
      case ETXTBSY:      code=G_FILE_ERROR_TXTBSY; break;
      case EFAULT:       code=G_FILE_ERROR_FAULT;  break;
      case ELOOP:        code=G_FILE_ERROR_LOOP;   break;
      case ENOSPC:       code=G_FILE_ERROR_NOSPC;  break;
      case ENOBUFS:
      case ENOMEM:       code=G_FILE_ERROR_NOMEM;  break;
      case EMFILE:       code=G_FILE_ERROR_MFILE;  break;
      case ENFILE:       code=G_FILE_ERROR_NFILE;  break;
      case EBADF:        code=G_FILE_ERROR_BADF;   break;

	 /* FIXME, declare error code for EPROTONOSUPPORT, EAFNOSUPPORT*/
      case EPROTONOSUPPORT:
      case EAFNOSUPPORT:
      case ENOTSOCK:
      case EINVAL:       code=G_FILE_ERROR_INVAL;  break;

      case EPIPE:        code=G_FILE_ERROR_PIPE;   break;
      case EAGAIN:     // not really error
	 *err = NULL;
	 return;
      case EINTR:        code=G_FILE_ERROR_INTR;   break;
      case EIO:          code=G_FILE_ERROR_IO;     break;
      case EPERM:        code=G_FILE_ERROR_PERM;   break;
//      case EFAILED:      code=G_FILE_ERROR_FAILED; break;
      default:
	 g_assert ( ! "unknown error code");
   }

   *err = g_error_new_literal (G_FILE_ERROR,
			       code,
			       strerror (errnum));
}

static 
gint oflags (gint flags)
{
	const gint ALLBITS 
	   = G_FILE_RDONLY
	   | G_FILE_WRONLY
	   | G_FILE_RDWR
	   | G_FILE_CREAT
	   | G_FILE_EXCL
	   | G_FILE_APPEND
	   | G_FILE_TRUNC;
	
	/* only known flags are accepted */
	g_assert ((flags & ~ALLBITS)==0);
	g_assert (flags == (O_RDONLY
			    | ((flags & G_FILE_WRONLY) ? O_WRONLY : 0)
			    | ((flags & G_FILE_RDWR)   ? O_RDWR   : 0)
			    | ((flags & G_FILE_CREAT)  ? O_CREAT  : 0)
			    | ((flags & G_FILE_EXCL)   ? O_EXCL   : 0)
			    | ((flags & G_FILE_APPEND) ? O_APPEND : 0)
			    | ((flags & G_FILE_TRUNC)  ? O_TRUNC  : 0)));

	return (O_RDONLY
		| ((flags & G_FILE_WRONLY) ? O_WRONLY : 0)
		| ((flags & G_FILE_RDWR)   ? O_RDWR   : 0)
		| ((flags & G_FILE_CREAT)  ? O_CREAT  : 0)
		| ((flags & G_FILE_EXCL)   ? O_EXCL   : 0)
		| ((flags & G_FILE_APPEND) ? O_APPEND : 0)
		| ((flags & G_FILE_TRUNC)  ? O_TRUNC  : 0));
	
	/* should be identical for linux */
}

static
gint 
sflags (gint flags)
{
	const gint ALLBITS =   G_FILE_IRUSR|G_FILE_IWUSR|G_FILE_IXUSR
                         | G_FILE_IRGRP|G_FILE_IWGRP|G_FILE_IXGRP
                         | G_FILE_IROTH|G_FILE_IWOTH|G_FILE_IXOTH;

	/* only known flags accepted */
	g_assert ((flags & ~ALLBITS)==0);

	g_assert (flags == (0
			    | ((flags & G_FILE_IRUSR) ? S_IRUSR : 0)
			    | ((flags & G_FILE_IWUSR) ? S_IWUSR : 0)
			    | ((flags & G_FILE_IXUSR) ? S_IXUSR : 0)

			    | ((flags & G_FILE_IRGRP) ? S_IRGRP : 0)
			    | ((flags & G_FILE_IWGRP) ? S_IWGRP : 0)
			    | ((flags & G_FILE_IXGRP) ? S_IXGRP : 0)

			    | ((flags & G_FILE_IROTH) ? S_IROTH : 0)
			    | ((flags & G_FILE_IWOTH) ? S_IWOTH : 0)
			    | ((flags & G_FILE_IXOTH) ? S_IXOTH : 0)));
	return  (0
		 | ((flags & G_FILE_IRUSR) ? S_IRUSR : 0)
		 | ((flags & G_FILE_IWUSR) ? S_IWUSR : 0)
		 | ((flags & G_FILE_IXUSR) ? S_IXUSR : 0)
		 
		 | ((flags & G_FILE_IRGRP) ? S_IRGRP : 0)
		 | ((flags & G_FILE_IWGRP) ? S_IWGRP : 0)
		 | ((flags & G_FILE_IXGRP) ? S_IXGRP : 0)
		 
		 | ((flags & G_FILE_IROTH) ? S_IROTH : 0)
		 | ((flags & G_FILE_IWOTH) ? S_IWOTH : 0)
		 | ((flags & G_FILE_IXOTH) ? S_IXOTH : 0));
	
}
gint 
g_getpid(GError** err)
{
   return (gint) getpid();
}

gint 
g_socket (gint     fam, 
	  gint     type, 
	  gint     prot, 
	  GError** err)
{
   const int res 
      = socket(fam, type, prot);

   if (res == -1 && err)
   {
      file_set_error (err, errno);
   }
   return res;
}

gint 
g_accept (gint             serverfd, 
	  struct sockaddr* addr, 
	  guint*           addrlen, 
	  GError**         err)
{
   const int res 
      = TEMP_FAILURE_RETRY(accept(serverfd, addr, (socklen_t*) addrlen));

   if (res == -1 && err)
   {
      file_set_error (err, errno);
   }
   return res;
}

gint g_bind (gint             sockfd, 
	     struct sockaddr* my_addr, 
	     size_t           addrlen, 
	     GError**         err)
{
   const int res 
      = bind(sockfd, my_addr, (socklen_t) addrlen);
   
   if (res == -1 && err)
   {
      file_set_error (err, errno);
   }
   return res;
}

gint    
g_connect (gint                   fd, 
	   const struct sockaddr* serv_addr, 
	   size_t                 addrlen, 
	   GError**               err)
{
   const int res 
      = TEMP_FAILURE_RETRY(connect (fd, serv_addr, (socklen_t) addrlen));

   if (res == -1 && err)
   {
      file_set_error (err, errno);
   }
   return res;
}

gint   
g_pipe (gint      fds[2], 
	GError**  err)
{
   const int res = pipe (fds);

   if (res == -1 && err)
   {
      file_set_error (err, errno);
   }

   if (res == -1)
   {
      fds[0] = -1;
      fds[1] = -1;
   }

   return res;
}

gint    
g_poll (GPollFD* ufds, 
	guint    nfds, 
	gint     timeout, 
	GError** err)
{
   // allocate memory on stack for n fds
   gint res = 0;
   struct pollfd * tmp_ufds 
      = (struct pollfd*) alloca (sizeof (struct pollfd) * nfds);
   
   bzero (tmp_ufds, sizeof (struct pollfd) * nfds);

   {
      gint i=0;
      for (i=0; i<nfds; ++i)
      {
	 tmp_ufds[i].fd = ufds[i].fd;
	 if (ufds[i].events &  G_IO_IN)  tmp_ufds[i].events |= POLLIN;
	 if (ufds[i].events &  G_IO_PRI) tmp_ufds[i].events |= POLLPRI;
	 if (ufds[i].events &  G_IO_OUT) tmp_ufds[i].events |= POLLOUT;
      }
   }
   
   res = TEMP_FAILURE_RETRY(poll (tmp_ufds, nfds, timeout));

   if (res == -1 && err)
   {
      file_set_error (err, errno);
      return -1;
   }

   {
      gint i = 0;
      for (i=0; i<nfds; ++i)
      {
	 ufds[i].revents = 0;  // reset
	 if (tmp_ufds[i].revents &  POLLIN)   ufds[i].revents |= G_IO_IN;
	 if (tmp_ufds[i].revents &  POLLOUT)  ufds[i].revents |= G_IO_OUT;
	 if (tmp_ufds[i].revents &  POLLPRI)  ufds[i].revents |= G_IO_PRI;
	 if (tmp_ufds[i].revents &  POLLERR)  ufds[i].revents |= G_IO_ERR;
	 if (tmp_ufds[i].revents &  POLLHUP)  ufds[i].revents |= G_IO_HUP;
	 if (tmp_ufds[i].revents &  POLLNVAL) ufds[i].revents |= G_IO_NVAL;
      }
   }
   return res;

   // _ufds will be remove from stack automaticly
}

gint
g_dup (gint  hd,
       gboolean inherit, 
       GError** err)
{
   int res = dup (hd);
   if (res == -1 && err)
   {
      file_set_error (err, errno);
   }
   else
   {
      if (!inherit)
      {
	 long cloexec = FD_CLOEXEC;
	 TEMP_FAILURE_RETRY(fcntl (res, F_SETFD, cloexec)); 

	 assert (TEMP_FAILURE_RETRY(fcntl (res, F_GETFD, 0)) == cloexec);
      }
   }
   return res;
}


gint 
g_ioctl (gint     fd, 
	 gint     cmd,
	 gpointer argp,
	 GError** err)
{
   switch (cmd)
   {
      case FIONBIO:
      {
	 gint res = TEMP_FAILURE_RETRY(ioctl ((int) fd,  (int) FIONBIO, argp)) ;
	 if (res==-1) 
	 {	
	     file_set_error (err, errno);
	    return -1;
	 }
	 return 0;
      }
      default:
	 g_assert ( ! "unknown command");
   }
}


/* this handle can't be used in context of poll() operation */
gint 
g_open (char*    file, 
	gint     flags,
	gint     mode,
	GError** err)
{
   int res = open (file, oflags (flags), sflags (mode));

   if (res == -1)
   {
       file_set_error (err, errno);
      return -1;
   }
   return res;
}

gint g_remove   (char*    file, 
		 GError** err)
{
   int res = remove (file);

   if (res == -1)
   {
       file_set_error (err, errno);
      return -1;
   }
   return res;
} 

gint 
g_listen  (gint     fd, 
	   guint    backlog, 
	   GError** err)
{
   int res = listen (fd, backlog);
   if (res == -1 && err)
   {
      file_set_error (err, errno);
   }
   return res;
}

gint 
g_shutdown (gint     fd, 
	    gint     how,
	    GError** err )
{
   if (shutdown(fd, how)==-1)
   {
       file_set_error (err, errno);
      return -1;
   }
   return 0;
}

gint 
g_close   (gint     fd, 
	   GError** err )
{
   int res = TEMP_FAILURE_RETRY(close (fd));
   if (res == -1 && err)
   {
      file_set_error (err, errno);
   }
   return res;
}



/* blocking and non-blocking mode */
gint 
g_send (gint       fd, 
	gconstpointer src, 
	size_t        len,
	gint          flags, 
	GError**      err)
{
   int res = TEMP_FAILURE_RETRY(send (fd, src, len, flags));
   if (res == -1 && err)
   {
      file_set_error (err, errno);
   }
   return res;
}

gint 
g_recv (gint   fd,
	gpointer  src, 
	size_t    len,
	gint      flags, 
	GError**  err )
{
   int res = TEMP_FAILURE_RETRY(recv (fd, src, len, flags));
   if (res == -1 && err)
   {
      file_set_error (err, errno);
   }
   return res;
}

/* blocking mode only */
gint 
g_write (gint          fd, 
	 gconstpointer src, 
	 size_t        len, 
	 GError**      err)
{
   int res = TEMP_FAILURE_RETRY(write (fd, src, len));
   if (res == -1 && err)
   {
      file_set_error (err, errno);
   }
   return res;
}

gint 
g_read (gint     fd, 
	gpointer src,
	size_t   len, 
	GError** err)
{
   int res = TEMP_FAILURE_RETRY(read (fd, src, len));
   if (res == -1 && err)
   {
      file_set_error (err, errno);
   }
   return res;
}


#define NANOS_PER_SEC ((gint)(1*1000*1000*1000))
#define MSECS_PER_LOOP ((gint)(50))

GQuark
g_process_error_quark (void)
{
   static GQuark quark = 0;
   if (quark == 0)
      quark = g_quark_from_static_string ("g-process-error-quark");
   return quark;
}

static 
void
process_set_error(GError** err, 
		  int      errnum)
{
   gint code = 0;
   if (err==NULL) return;
   switch (errnum)
   {
//      case EFORK:  code=G_PROCESS_ERROR_FORK; break;   /* fork failed due to lack of memory */
//      case EREAD:  code=G_PROCESS_ERROR_READ; break;   /* read or select on pipes failed */
//      case ECHDIR: code=G_PROCESS_ERROR_CHDIR; break;  /* changing to working dir failed */
      case EACCES: code=G_PROCESS_ERROR_ACCES; break;  /* execv() returned EACCES */
      case EPERM:  code=G_PROCESS_ERROR_PERM; break;   /* execv() returned EPERM */
      case E2BIG:  code=G_PROCESS_ERROR_2BIG; break;   /* execv() returned E2BIG */
      case ENOEXEC:code=G_PROCESS_ERROR_NOEXEC; break; /* execv() returned ENOEXEC */
      case ENAMETOOLONG: code=G_PROCESS_ERROR_NAMETOOLONG; break; /* ""  "" ENAMETOOLONG */
      case ENOENT: code=G_PROCESS_ERROR_NOENT; break;       /* ""  "" ENOENT */
      case ENOMEM: code=G_PROCESS_ERROR_NOMEM; break;       /* ""  "" ENOMEM */
      case ENOTDIR:code=G_PROCESS_ERROR_NOTDIR; break;      /* ""  "" ENOTDIR */
      case ELOOP:  code=G_PROCESS_ERROR_LOOP; break;        /* ""  "" ELOOP   */
//      case ETXTBUSY: code=G_PROCESS_ERROR_TXTBUSY; break;     /* ""  "" ETXTBUSY */
      case EIO:    code=G_PROCESS_ERROR_IO; break;          /* ""  "" EIO */
      case ENFILE: code=G_PROCESS_ERROR_NFILE; break;       /* ""  "" ENFILE */
      case EMFILE: code=G_PROCESS_ERROR_MFILE; break;       /* ""  "" EMFLE */
      case EINVAL: code=G_PROCESS_ERROR_INVAL; break;       /* ""  "" EINVAL */
      case EISDIR: code=G_PROCESS_ERROR_ISDIR; break;       /* ""  "" EISDIR */
      case ELIBBAD:code=G_PROCESS_ERROR_LIBBAD; break;      /* ""  "" ELIBBAD */
/*      case EFAILED:code=G_PROCESS_ERROR_FAILED; break; */   /* other fatal failure, error->message
							     * should explain
							     */
      default:
	 g_assert ( ! "unknown error code");
   }
   *err = g_error_new_literal (G_PROCESS_ERROR,
			       code,
			       strerror (errnum));
}

gint
g_sleep (gint msec, GError** err)
{
   struct timespec delay  = {0, 0};
   struct timespec remain = {0, 0};

   delay.tv_sec  = (msec) / 1000;
   delay.tv_nsec = ((msec) % 1000)*1000*1000; 
   
/*  g_print ("%d/%d", delay.tv_sec, delay.tv_nsec); */

   while (delay.tv_sec+delay.tv_nsec>0)
   {
      if (nanosleep (&delay, &remain)==-1)
      {
	 if (errno==EINTR)
	 {
	    delay = remain;
	 }
	 else
	 {
	    process_set_error (err, errno);
	    return -1;
	 }
      }
      else
      {
	 return 0;
      }
   }
}

gproc
g_process_new (char* const  argv[],
	       gint      fdin,
	       gint      fdout,
	       gint      fderr,
	       guint     flags,
	       GError**  err)
{
   pid_t pid = -1;
   
   // g_assert (file!=NULL || argv!=NULL);

   switch(pid = fork()) 
   {
      case 0:	/* the child */
	 if (fdin!=-1) 
	 {
	    // at first close, then copy to 0
	    if (dup2(fdin, 0)==-1)
	    {
	       exit (errno);
	    }
	 }
	 if (fdout!=-1) 
	 { 
	    // at first close, then copy to 1
	    if (dup2(fdout, 1)==-1)
	    {
	       exit (errno);
	    }
	 }
	 if (fderr!=-1) 
	 {
	    // at first close, then copy to 2
	    if (dup2(fderr, 2)==-1)
	    {
	       exit (errno);
	    }
	 }

// #ifndef sun
	 if(flags & G_PROCESS_DETACH) {
	    int fd = open("/dev/tty", O_RDONLY);
	    if(fd != -1) {
	       if(ioctl(fd, TIOCNOTTY) < 0) 
	       {
		  exit (errno);
	       }
	       close(fd);
	    }
	    else 
	    {
	       exit (errno);
	    }
	 }
// #endif

	 if(flags & G_PROCESS_NEW_GROUP)
	 {
/* #ifdef sun */
/* 	    if (setpgrp() < 0)  */
/* #else */
	    if (setpgrp() != 0)
/* #endif */
	    {
	       exit (errno);
	    }
	 }

	 if (TEMP_FAILURE_RETRY(execvp(argv[0], (char *const*)argv))==-1)
	 {
	    exit (errno);
	 }
	 // never returns
	 return 0;

      case -1:
	 process_set_error (err, errno);
	 return 0;

      default:/* the parent */
	 return pid;
   }
}

gint 
g_process_kill (gproc proc, 
		gint sig, 
		GError** err)
{

   if (!(sig & G_SIGTERM) && !(sig & G_SIGINT) && !(sig & G_SIGKILL)) 
      g_warning ("non portable signal (%d) used, only"
		 " G_SIGTERM, G_SIGKILL, and G_SIGINT are supported on win32", sig);

   if (kill ((int) proc, sig)==-1)
   {
      process_set_error (err, errno);
      return -1;
   }
   return 0;
}

gint 
g_process_wait (gproc     proc, 
		gint*    status,
		gint     timeout,
		GError** err)
{
   gint TIME_UNIT=10;      /* msecs */
   gint elapsed_time=0;  /* msecs */

   for (elapsed_time=0; 
	elapsed_time<timeout || timeout<=0 ; 
	elapsed_time+=TIME_UNIT)
   {
      gint exit_status=0;   
      switch (waitpid ((int) proc, &exit_status, WNOHANG))
      {
	 case -1:
	    if (errno!=EINTR)
	    {
	       process_set_error (err, errno);
	       return -1;
	    }
	    break;
	 case 0:
	    if (timeout==0) 
	       return 0;
	    else
	       usleep(TIME_UNIT*1000);
	    break;
	 default:
	    if (status) *status = WEXITSTATUS(exit_status);
	    return proc;
      }
   }
   
   return 0;
}

gint 
g_process_free (gproc proc,
		GError** err)
{
   /* this function is only needed for win32 */
   return 0;
}

#endif /* WIN32 */

