/* GIOS - General IO set
 * Copyright (C) 2002  Frank Rehberger, Frank.Rehberger@web.de
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#ifndef __G_IOS_H__
#define __G_IOS_H__

#ifdef WIN32

#include <windows.h>
#define G_STDIN  ((gint)GetStdHandle(STD_INPUT_HANDLE))
#define G_STDOUT ((gint)GetStdHandle(STD_OUTPUT_HANDLE))
#define G_STDERR ((gint)GetStdHandle(STD_ERROR_HANDLE))

#else

#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <asm/ioctls.h>
#define G_STDIN  ((gint)0)
#define G_STDOUT ((gint)1)
#define G_STDERR ((gint)2)

#endif

#define G_WAIT_NOHANG          01    /* for use with g_waitpid */
 
#define __G_FILE_IREAD       0400    /* Read by owner.  */
#define __G_FILE_IWRITE      0200    /* Write by owner.  */
#define __G_FILE_IEXEC       0100    /* Execute by owner.  */

#define G_FILE_IRUSR __G_FILE_IREAD       /* Read by owner.  */
#define G_FILE_IWUSR __G_FILE_IWRITE      /* Write by owner.  */
#define G_FILE_IXUSR __G_FILE_IEXEC       /* Execute by owner.  */
/* Read, write, and execute by owner.  */
#define G_FILE_IRWXU (__G_FILE_IREAD|__G_FILE_IWRITE|__G_FILE_IEXEC)

# define G_FILE_IREAD        G_FILE_IRUSR
# define G_FILE_IWRITE       G_FILE_IWUSR
# define G_FILE_IEXEC        G_FILE_IXUSR

#define G_FILE_IRGRP (G_FILE_IRUSR >> 3)  /* Read by group.  */
#define G_FILE_IWGRP (G_FILE_IWUSR >> 3)  /* Write by group.  */
#define G_FILE_IXGRP (G_FILE_IXUSR >> 3)  /* Execute by group.  */
/* Read, write, and execute by group.  */
#define G_IRWXG (G_IRWXU >> 3)

#define G_FILE_IROTH (G_FILE_IRGRP >> 3)  /* Read by others.  */
#define G_FILE_IWOTH (G_FILE_IWGRP >> 3)  /* Write by others.  */
#define G_FILE_IXOTH (G_FILE_IXGRP >> 3)  /* Execute by others.  */

/* from /usr/include/asm/fcntl.h */
#define G_FILE_RDONLY      00
#define G_FILE_WRONLY      01
#define G_FILE_RDWR        02
#define G_FILE_CREAT     0100 
#define G_FILE_EXCL      0200 
/*#define G_FILE_NOCTTY    0400*/
#define G_FILE_TRUNC     01000 
#define G_FILE_APPEND   02000

/* currently SIGINT and SIGTERM not supported */
#define G_SIGINT    2  /* currently not supported on win32 */
#define G_SIGKILL   9 
#define G_SIGTERM  15  /* currently not supported on win32 */


#include <glib.h>

#define G_IO_NONBLOCKING(fd, eptr)               \
    { gint arg=1; gint res=g_ioctl (fd, FIONBIO, (gpointer)&arg, eptr); res;}

#define G_IO_BLOCKING(fd, eptr)           \
    { gint arg=0; gint res=g_ioctl (fd, FIONBIO, (gpointer)&arg, eptr); res;}                       

gint g_ios_init (GError** err);

/** */
gint g_getpid   (GError** err);

/** */
gint g_socket   (gint fam, 
		 gint type,
		 gint prot, 
		 GError** err);

gint g_bind     (gint             sockfd, 
		 struct sockaddr* my_addr, 
		 size_t           addrlen, 
		 GError**         err);

gint g_accept   (gint             fd, 
		 struct sockaddr* addr, 
		 guint*           addrlen, 
		 GError**         err);

gint g_connect  (gint             fd, 
		 const struct sockaddr *serv_addr, 
		 size_t addrlen,
		 GError** err);

gint g_listen   (gint  fd, 
		 guint    queuesize,
		 GError** err);

gint g_close    (gint fd, 
		 GError** err);

gint g_shutdown (gint fd, 
		 gint how,
		 GError** err);

gint g_pipe     (gint     fds[2],
		 GError** err);

gint g_poll     (GPollFD*     ufds, 
		 guint        nfds, 
		 gint         timeout, 
		 GError**     err); 

gint g_dup      (gint  fd, 
		 gboolean inherit, 
		 GError** err);

gint g_ioctl    (gint     fd, 
				 gint     cmd, 
				 gpointer argp, 
				 GError** err);

/* blocking and non-blocking mode */
gint g_send     (gint fd, 
		 gconstpointer src, 
		 size_t len, 
		 gint flags, 
		 GError** err);

gint g_recv     (gint fd, 
		 gpointer src,
		 size_t len, 
		 gint flags,
		 GError **err);

/* this handle can't be used in context of poll() operation */
gint g_open     (char*    file, 
		 gint     flags, 
		 gint     mode,
		 GError** err); 

gint g_remove   (char*    file, 
		 GError** err); 

/* blocking mode only */
gint g_write    (gint fd, 
		 gconstpointer src, 
		 size_t len,
		 GError **err);

gint g_read     (gint fd, 
		 gpointer src, 
		 size_t len,
		 GError **err);

gint g_sleep    (gint msec, GError** err);


#define G_PROCESS_ERROR g_process_error_quark ()

typedef enum
{ 
   G_PROCESS_ERROR_FORK,
   G_PROCESS_ERROR_READ,
   G_PROCESS_ERROR_ACCES,
   G_PROCESS_ERROR_PERM,
   G_PROCESS_ERROR_2BIG,
   G_PROCESS_ERROR_NOEXEC,
   G_PROCESS_ERROR_NAMETOOLONG, 
   G_PROCESS_ERROR_NOENT,
   G_PROCESS_ERROR_NOMEM, 
   G_PROCESS_ERROR_NOTDIR, 
   G_PROCESS_ERROR_LOOP, 
   G_PROCESS_ERROR_TXTBUSY, 
   G_PROCESS_ERROR_IO, 
   G_PROCESS_ERROR_NFILE, 
   G_PROCESS_ERROR_MFILE, 
   G_PROCESS_ERROR_INVAL, 
   G_PROCESS_ERROR_ISDIR, 
   G_PROCESS_ERROR_LIBBAD, 
   G_PROCESS_ERROR_FAILED
} GProcessError;

#define G_PROCESS_DETACH            0x0001
#define G_PROCESS_NEW_GROUP         0x0002

typedef gint gproc;

gproc 
g_process_new (char* const  argv[], 
	       gint         fdin,
	       gint         fdout,
	       gint         fderr,
	       guint        flags,
	       GError**     err);

gint
g_process_kill   (gproc prc, 
		  gint sig, 
		  GError** err);

gint
g_process_wait   (gproc prc, 
	          gint* stat, 
	          gint  msecs,
	          GError** err);

gint 
g_process_free (gproc prc,
		GError **err);

#endif
