/* GIOS - General IO set
 * Copyright (C) 2002  Frank Rehberger, Frank.Rehberger@web.de
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */


#ifdef WIN32 

#include <windows.h>
#include <winsock2.h>
#include <errno.h>
#include <assert.h>
#include <fcntl.h>
#include <stdarg.h>
#include <signal.h>
#include <sys/stat.h>

/* #include <winsock2.h> */
/* #include <ioctl.h> */
#include "gios.h"

gint 
tcp_pipe (gint fds[2], GError **err);


static 
void
file_set_error(GError** err, 
	       int      errnum)
{
   gint code = 0;
  
   if (err==NULL) return;
   /* g_warning ("errcode=%d\n", errnum); */

   switch (errnum)
   {
      // FIXME, set WSA-error codes
      case EEXIST:       code=G_FILE_ERROR_EXIST;  break;
      case EISDIR:       code=G_FILE_ERROR_ISDIR;  break;
      case WSAEACCES:
      case EACCES:       code=G_FILE_ERROR_ACCES;  break;
      case ENAMETOOLONG: code=G_FILE_ERROR_NAMETOOLONG; break;
      case ENOENT:       code=G_FILE_ERROR_NOENT;  break;
      case ENOTDIR:      code=G_FILE_ERROR_NOTDIR; break;
      case ENXIO:        code=G_FILE_ERROR_NXIO;   break;
      case ENODEV:       code=G_FILE_ERROR_NODEV;  break;
      case EROFS:        code=G_FILE_ERROR_ROFS;   break;
	 /* case ETXTBSY:      code=G_FILE_ERROR_TXTBSY; break; */
      case EFAULT:       code=G_FILE_ERROR_FAULT;  break;
	 /* case ELOOP:        code=G_FILE_ERROR_LOOP;   break; */
      case ENOSPC:       code=G_FILE_ERROR_NOSPC;  break;
      case WSAENOBUFS:
	 /* case ENOBUFS:*/
      case ENOMEM:       code=G_FILE_ERROR_NOMEM;  break;
      case WSAEMFILE:
      case EMFILE:       code=G_FILE_ERROR_MFILE;  break;
      case ENFILE:       code=G_FILE_ERROR_NFILE;  break;
      case WSAENOTSOCK:
      case EBADF:        code=G_FILE_ERROR_BADF;   break;

	 /* FIXME, declare error code for EPROTONOSUPPORT, EAFNOSUPPORT */
      case WSAEADDRINUSE:     

      case WSAEPROTONOSUPPORT:  
      case WSAEPROTOTYPE:       
      case WSAESOCKTNOSUPPORT:   
      case WSAEINVAL:          
      case WSAEADDRNOTAVAIL:   
	 /* case EPROTONOSUPPORT: */
	 /* case EAFNOSUPPORT:    */
      case 87:                  /* ERROR_INVALID_PARAMETER */           
      case EINVAL:       code=G_FILE_ERROR_INVAL;  break;

      case EPIPE:        code=G_FILE_ERROR_PIPE;   break;

      case WSAEALREADY:	    /* TESTME, validate mapping */
      case WSAEWOULDBLOCK:  /* same as EAGAIN */
      case EAGAIN:     // not really error
	 *err = NULL;
	 return;
      case EINTR:        code=G_FILE_ERROR_INTR;   break;
      case EIO:          code=G_FILE_ERROR_IO;     break;
      case EPERM:        code=G_FILE_ERROR_PERM;   break;

      case WSAETIMEDOUT:          
      case WSAENETUNREACH:     
      case WSAENETDOWN:              
      case WSAEAFNOSUPPORT:    
      case WSAECONNREFUSED:     
      case WSAECONNABORTED:    
      case WSAEOPNOTSUPP:      
      case WSAEISCONN:         
      case WSAESHUTDOWN:       
      case WSAENETRESET:        
      case WSAENOTCONN:       
      case WSAEMSGSIZE:       
      case WSAEHOSTUNREACH:   
	 /* case EFAILED: */    code=G_FILE_ERROR_FAILED;         break; 
	  
      case WSANOTINITIALISED:
	 g_error ("network not initalized, to be called WSAStartup(DWORD,DWORD)");
	 exit (1);
      case WSAEINPROGRESS:
      case WSAEFAULT:
      case WSAEINTR:
	 g_error ("winsock1.1 specific error codes");
	 break;
      default:
	 g_warning ("unknown errcode=%d\n", errnum);
	 code=G_FILE_ERROR_FAILED;
   }

   *err = g_error_new_literal (G_FILE_ERROR,
			       code,
			       strerror (errnum));
}

/* FIXME, dynamic loading of WinSock2 DLL (ws2) */ 
static 
gint 
init_winsock (GError** err)
{
   static first_call = TRUE;
   if (first_call)
   {
      first_call = FALSE;

      {
	 WORD version;
	 WSADATA data;
	 int stat;
	 
	 version = MAKEWORD( 2, 2 );
	 
	 stat = WSAStartup( version, &data );
	 if ( stat != 0 ) {
	    g_file_set_error (err, WSAGetLastError());
	    return -1;
	 }
	 
	 /* Confirm that the WinSock DLL supports 2.2.*/
	 /* Note that if the DLL supports versions greater    */
	 /* than 2.2 in addition to 2.2, it will still return */
	 /* 2.2 in wVersion since that is the version we  */
	 /* requested.                                        */
	 
	 if ( LOBYTE( data.wVersion ) != 2 ||
	      HIBYTE( data.wVersion ) != 2 ) {
	    WSACleanup( );
	    g_file_set_error (err, WSAGetLastError());/*FIXME,wrong error code*/
	    return -1;
	 }
	 /* The WinSock DLL is acceptable. Proceed. */
      }
   }
   return 0;
}

gint 
g_ios_init (GError** err)
{
   return init_winsock (err);
}


gint
g_socket (gint fam, gint type, gint prot, 
	  GError** err)
{
   if (init_winsock(err)==-1) return -1;
   {
      SOCKET fd = socket(fam, type, prot);
      if (fd == INVALID_SOCKET)
      {
	 /*
	   WSANOTINITIALISED A successful WSAStartup call must occur
	   before using this function.  WSAENETDOWN The network
	   subsystem or the associated service provider has failed.
	   WSAEAFNOSUPPORT The specified address family is not
	   supported.  WSAEINPROGRESS A blocking Windows Sockets 1.1
	   call is in progress, or the service provider is still
	   processing a callback function.  WSAEMFILE No more socket
	   descriptors are available.  WSAENOBUFS No buffer space is
	   available. The socket cannot be created.
	   WSAEPROTONOSUPPORT The specified protocol is not supported.
	   WSAEPROTOTYPE The specified protocol is the wrong type for
	   this socket.  WSAESOCKTNOSUPPORT The specified socket type
	   is not supported in this address family.WSAETIMEDOUT.
	 */
	  file_set_error (err, WSAGetLastError ());
	 return (gint) -1;
      }
      return (gint) fd;
   }
}

gint
g_accept (gint             serverfd, 
	  struct sockaddr* addr, 
	  size_t*          addrlen, 
	  GError**         err)
{
   init_winsock ();
   {
      SOCKET fd = accept(serverfd, addr, addrlen);
      if (fd == INVALID_SOCKET)
      {
	 /*
	   WSANOTINITIALISED A successful WSAStartup call must occur
	   before using this function.  WSAENETDOWN The network
	   subsystem has failed.  WSAEFAULT The addrlen parameter is
	   too small or addr is not a valid part of the user address
	   space.  WSAEINTR A blocking Windows Sockets 1.1 call was
	   canceled through WSACancelBlockingCall.  WSAEINPROGRESS A
	   blocking Windows Sockets 1.1 call is in progress, or the
	   service provider is still processing a callback function.
	   WSAEINVAL The listen function was not invoked prior to
	   accept.  WSAEMFILE The queue is nonempty upon entry to
	   accept and there are no descriptors available.  WSAENOBUFS
	   No buffer space is available.  WSAENOTSOCK The descriptor
	   is not a socket.  WSAEOPNOTSUPP The referenced socket is
	   not a type that supports connection-oriented service.
	   WSAEWOULDBLOCK The socket is marked as nonblocking and no
	   connections are present to be accepted.
	 */
	  file_set_error (err, WSAGetLastError ());
	 return -1;
      }
      return (gint) fd;
   }
}

gint    
g_connect (gint                   fd, 
	   const struct sockaddr* serv_addr, 
	   size_t                 addrlen, 
	   GError**               err)
{
   if (init_winsock(err)==-1) return -1;
   {
      int res = connect (fd, serv_addr, addrlen);
      if (res == SOCKET_ERROR)
      {
	 /*
	   WSANOTINITIALISED A successful WSAStartup call must occur
	   before using this function.  WSAENETDOWN The network
	   subsystem has failed.  WSAEADDRINUSE The socket's local
	   address is already in use and the socket was not marked to
	   allow address reuse with SO_REUSEADDR. This error usually
	   occurs when executing bind, but could be delayed until this
	   function if the bind was to a partially wildcard address
	   (involving ADDR_ANY) and if a specific address needs to be
	   committed at the time of this function.  WSAEINTR The
	   blocking Windows Socket 1.1 call was canceled through
	   WSACancelBlockingCall.  WSAEINPROGRESS A blocking Windows
	   Sockets 1.1 call is in progress, or the service provider is
	   still processing a callback function.  WSAEALREADY A
	   nonblocking connect call is in progress on the specified
	   socket.

	   Note In order to preserve backward compatibility, this
	   error is reported as WSAEINVAL to Windows Sockets 1.1
	   applications that link to either WINSOCK.DLL or
	   WSOCK32.DLL.  WSAEADDRNOTAVAIL The remote address is not a
	   valid address (such as ADDR_ANY).  WSAEAFNOSUPPORT
	   Addresses in the specified family cannot be used with this
	   socket.  WSAECONNREFUSED The attempt to connect was
	   forcefully rejected.  WSAEFAULT The name or the namelen
	   parameter is not a valid part of the user address space,
	   the namelen parameter is too small, or the name parameter
	   contains incorrect address format for the associated
	   address family.  WSAEINVAL The parameter s is a listening
	   socket.  WSAEISCONN The socket is already connected
	   (connection-oriented sockets only).  WSAENETUNREACH The
	   network cannot be reached from this host at this time.
	   WSAENOBUFS No buffer space is available. The socket cannot
	   be connected.  WSAENOTSOCK The descriptor is not a socket.
	   WSAETIMEDOUT Attempt to connect timed out without
	   establishing a connection.  WSAEWOULDBLOCK The socket is
	   marked as nonblocking and the connection cannot be
	   completed immediately.  WSAEACCES Attempt to connect
	   datagram socket to broadcast address failed because
	   setsockopt option SO_BROADCAST is not enabled.
	 */
	  file_set_error (err, WSAGetLastError ());
	 return (gint) -1;
      }
      
      return (gint) 0;
   }
}

gint   
g_pipe (gint fds[2], GError** err)
{
   init_winsock ();
   /* return CreatePipe(&fds[0], &fds[1], NULL, 0) ? 0 : -1; */

   /** Note: because windows native pipes can not be used for "select" operations, for 
       portability reasons tcp-pipes are created here.
   */
   return tcp_pipe (fds, err); 
}

/* typedef struct fd_set { */
/*   u_int fd_count; */
/*   SOCKET fd_array[FD_SETSIZE]; */
/* } fd_set; */

typedef struct 
{
      u_int fd_count; 
      SOCKET fd_array[1];
} fd_set_head;

gint    
g_poll (GPollFD *ufds, guint nfds, gint timeout, GError** err)
{
   if (init_winsock(err)==-1) return -1;
   {
      /* "select" on windows can not handle sum of 0 file descriptors */
      if (nfds==0) 
      {
	 if (timeout>=0) 
	    Sleep (timeout);
	 else
	    Sleep (INFINITE);
	 
	 return 0;
      }
      else
      {
	 int fd_setsize = sizeof(fd_set_head)+(nfds*sizeof(SOCKET));
	 int i=0;
	 int res=0;
	 struct timeval tv[1];
	 
	 /* allocate memory on stack */
	 FD_SET*   rfds = (FD_SET*) _alloca (fd_setsize);
	 FD_SET*   wfds = (FD_SET*) _alloca (fd_setsize);
	 FD_SET*   efds = (FD_SET*) _alloca (fd_setsize);
	 
	 FD_ZERO (rfds);
	 FD_ZERO (wfds);
	 FD_ZERO (efds);
	 
	 /* transform msec to [sec,usec] */
	 tv->tv_sec  = timeout / 1000;
	 tv->tv_usec = timeout % 1000;
	 
	 /* transform ufds-list to fd_sets */
	 for (i=0; i<(int)nfds; ++i)
	 {
	    if (ufds[i].events &  G_IO_IN)  FD_SET (ufds[i].fd, rfds);
	    if (ufds[i].events &  G_IO_PRI) FD_SET (ufds[i].fd, rfds);
	    if (ufds[i].events &  G_IO_OUT) FD_SET (ufds[i].fd, wfds);
	    FD_SET (ufds[i].fd, efds);
	 }
	 
	 /** 
	     Note: On windows platform "select" fails if sum of file descriptors is 0. 
	     This behaviour is not POSIX compliant!  This means, you can not use "select" 
	     to implement platform portable microsecond sleep :(
	 */
	 res=select (0,     /* unused for win32 */
		     rfds,
		     wfds,
		     efds,
		     tv);
	 
	 if (res==0) 
	    /* timeout occured */
	    return 0;
	 
	 if (res==SOCKET_ERROR)
	 {
	    /*
	      WSANOTINITIALISED A successful WSAStartup call must
	      occur before using this function.  WSAEFAULT The Windows
	      Sockets implementation was unable to allocate needed
	      resources for its internal operations, or the readfds,
	      writefds, exceptfds, or timeval parameters are not part
	      of the user address space.  WSAENETDOWN The network
	      subsystem has failed.  WSAEINVAL The time-out value is
	      not valid, or all three descriptor parameters were NULL.
	      WSAEINTR A blocking Windows Socket 1.1 call was canceled
	      through WSACancelBlockingCall.  WSAEINPROGRESS A
	      blocking Windows Sockets 1.1 call is in progress, or the
	      service provider is still processing a callback
	      function.  WSAENOTSOCK One of the descriptor sets
	      contains an entry that is not a socket.
	    */
	    
	    // invalid values, return error value 
	     file_set_error (err, WSAGetLastError ());
	    return -1;
	 }
	 
	 assert (res>0);
	 
	 /* Iterate all given fds and in case an event occured set return
	  * event flags and remove from fd_set, otherwise mapping from
	  * unique fd_set to ufds-list might multiply return events. 
	  
	  * OPTME: this has O(n^2) in worst case, enhancement by
	  * index-lists
	  */
	 for (i=0; i<(int)nfds; ++i) 
	 {
	    /* reset return events */
	    ufds[i].revents=0;
	    
	    if (FD_ISSET ((SOCKET) (ufds[i].fd), rfds))
	    {
	       ufds[i].revents |= G_IO_IN;
	       FD_CLR ((SOCKET) (ufds[i].fd), rfds);
	    }
	    
	    if (FD_ISSET ((SOCKET) (ufds[i].fd), wfds))
	    {
	       ufds[i].revents |= G_IO_OUT;
	       FD_CLR ((SOCKET) (ufds[i].fd), wfds);
	    }
	    
	    if (FD_ISSET ((SOCKET) (ufds[i]).fd, efds))
	    {
	       /* FIXME, check for HUP event */
	       ufds[i].revents |= G_IO_ERR;
	       FD_CLR ((SOCKET) (ufds[i].fd), efds);
	    }
	 }
	 
	 return res;
      }
   }
}

gint
g_dup (gint fd, gboolean inherit, GError** err)
{
   init_winsock ();
   {
      /* Note: you might think this could be done using libc-file-handles, too. 
	 In this case your code would look like: <b><b>

	 orig=get_oshandle(fd);<b>
	 if (!DuplicateHandle(_, orig, _ &copy,...))...<b>
	 return open_osfhandle (copy, "ar");<b><b>

	 The last operation "open_osfhandle" will allocate resources that 
	 can not be freed and after more or less 2000 calls your system 
	 will run out of FileHandles. Cause of this, avoid libc-filehandles 
	 on windows platforms.
      */


      HANDLE orig = (HANDLE) fd;
      HANDLE copy = INVALID_SOCKET;
      
      if (!DuplicateHandle (GetCurrentProcess(), orig,
			    GetCurrentProcess(), &copy,
			    0,                   // parameter ignored,
			    (BOOL) inherit,      // inherit?
			    DUPLICATE_SAME_ACCESS))
      {
	 /*
	  */
	  file_set_error (err, GetLastError ());
	 return -1;
      }
      
      return (gint) copy;
   }
}

gint g_remove   (char*    filename, 
		 GError** err)
{
   if ( ! DeleteFile (filename))
   {
       file_set_error (err, GetLastError ());
      return -1;
   }
   return 0;
}

/* FIXME, use varargs */
/* this handle can't be used in context of poll() operation */
gint
g_open (char* file, 
	gint flags, 
	gint mode, /* FIXME, currently unused */
	GError** err)
{
   /** NOte: Only files opened with CreateFile operation can be used in shared mode. 
       Exclusive access is default on win platform, in oposite to posix compliant platforms */
   if (init_winsock(err)==-1) return -1;
   {
      DWORD access = 0;
      DWORD shmode = 0;
      DWORD dispo  = 0;
      
      SECURITY_ATTRIBUTES sec;
      ZeroMemory(&sec, sizeof(sec));
      sec.nLength              = sizeof(sec);
      sec.lpSecurityDescriptor = NULL;
      sec.bInheritHandle       = TRUE;
      
      if ((flags & G_FILE_APPEND) == G_FILE_APPEND)
      {
	 access |= FILE_APPEND_DATA;
      }
      /* G_FILE_DIRECTORY not supported for win32 */
      /*  
	  else if (flags & G_FILE_DIRECTORY)
	  {
	  flags |= FILE_ADD_SUBDIRECTORY;
	  } 
      */
      else
      {
	 if ((flags & G_FILE_RDONLY) == G_FILE_RDONLY) 
			access |= GENERIC_READ;
	 if ((flags & G_FILE_WRONLY) == G_FILE_WRONLY) 
			access |= GENERIC_WRITE;
	 if ((flags & G_FILE_RDWR) == G_FILE_RDWR)   
			access |= GENERIC_READ | GENERIC_WRITE;
      }
      
      // for unix, shared mode is default
      if ( ! ((flags & G_FILE_EXCL) == G_FILE_EXCL)) 
			shmode |= FILE_SHARE_READ | FILE_SHARE_WRITE;
      
      if ((flags & G_FILE_CREAT) == G_FILE_CREAT)  
      {
	 if ((flags & G_FILE_TRUNC) == G_FILE_TRUNC)   
		dispo = CREATE_ALWAYS;   
	 else                  
		dispo = OPEN_ALWAYS;
      }
      else
      {
	 if ((flags & G_FILE_TRUNC) == G_FILE_TRUNC)   
		dispo = TRUNCATE_EXISTING;  
	 else                   
		 dispo = OPEN_EXISTING;
      }

      {
	 HANDLE fd = CreateFile (file,
				 access, // read/write access
				 shmode, // sharing mode
				 &sec,   // security attribute
				 dispo,  // open mode
				 FILE_ATTRIBUTE_NORMAL,  // normal file 
				 NULL);
      
	 if (fd == INVALID_HANDLE_VALUE)
	 {
	     file_set_error (err, GetLastError ());
	    return -1;
	 }
      
	 return (gint) fd;
      }
   }
}

gint 
g_listen  (gint fd, 
	   guint backlog, 
	   GError** err)
{
   if (init_winsock(err)==-1) return -1;
   {
      int res = listen (fd, backlog);
      if (res == SOCKET_ERROR)
      {
	 /*
	   WSANOTINITIALISED A successful WSAStartup call must occur
	   before using this function.  WSAENETDOWN The network
	   subsystem has failed.  WSAEADDRINUSE The socket's local
	   address is already in use and the socket was not marked to
	   allow address reuse with SO_REUSEADDR. This error usually
	   occurs during execution of the bind function, but could be
	   delayed until this function if the bind was to a partially
	   wildcard address (involving ADDR_ANY) and if a specific
	   address needs to be committed at the time of this function.
	   WSAEINPROGRESS A blocking Windows Sockets 1.1 call is in
	   progress, or the service provider is still processing a
	   callback function.  WSAEINVAL The socket has not been bound
	   with bind.  WSAEISCONN The socket is already connected.
	   WSAEMFILE No more socket descriptors are available.
	   WSAENOBUFS No buffer space is available.  WSAENOTSOCK The
	   descriptor is not a socket.  WSAEOPNOTSUPP The referenced
	   socket is not of a type that supports the listen operation.
	 */
	  file_set_error (err, GetLastError());
	 return -1;
      }

      return res;
   }
}

gint 
g_shutdown  (gint fd, 
	     gint how,
	     GError** err )
{
   init_winsock ();

   if (shutdown((SOCKET)fd, how)==SOCKET_ERROR)
   {
      /*
	WSANOTINITIALISED A successful WSAStartup call must occur before using this function. 
	WSAENETDOWN The network subsystem has failed. 
	WSAEINVAL The how parameter is not valid, or is not consistent with the socket type. For example, SD_SEND is used with a UNI_RECV socket type. 
	WSAEINPROGRESS A blocking Windows Sockets 1.1 call is in progress, or the service provider is still processing a callback function. 
	WSAENOTCONN The socket is not connected (connection-oriented sockets only). 
	WSAENOTSOCK The descriptor is not a socket. 
      */
       file_set_error (err, WSAGetLastError());
      return -1;
   }
   return 0;
}

gint 
g_close   (gint fd, 
	   GError** err )
{
   init_winsock ();

   /* Note: if MSDN docmuentation is correct, "CloseHandle" can 
      be called on whatever kernel resource, too, sokets. Somehow, I observed some
      handle-leak on win2000 for sockets using CloseHandle, so I stuck to closesocket, */
   if (!CloseHandle((HANDLE)fd))
   {
       file_set_error (err, GetLastError());
      return -1;
   }
   return 0;

   /*
     else if (getsockopt (fd, SOL_SOCKET, SO_TYPE, NULL, NULL)!=SO_ERROR)
     {
     if (closesocket (fd)==SOCKET_ERROR)
     {
   */	    /**
	       WSANOTINITIALISED A successful WSAStartup call must occur
	       before using this function.
	 
	       WSAENETDOWN The network subsystem has failed.
	    
	       WSAENOTSOCK The descriptor is not a socket.
	    
	       WSAEINPROGRESS A blocking Windows Sockets 1.1 call is
	       in progress, or the service provider is still
	       processing a callback function.  
	    
	       WSAEINTR The (blocking) Windows Socket 1.1 call was
	       canceled through WSACancelBlockingCall.
	    
	       WSAEWOULDBLOCK The socket is marked as nonblocking and
	       SO_LINGER is set to a nonzero time-out value.
	    */
/*		  file_set_error (err, WSAGetLastError());
		 return -1;
		 }
		 return 0;
		 }
		 else
		 {
		 if (!CloseHandle((HANDLE)fd))
		 {
		  file_set_error (err, GetLastError());
		 return -1;
		 }
		 return 0;
		 } */
}


static
gint
_g_write (gint fd, 
	  gconstpointer src, 
	  size_t len, 
	  GError** err)
{
   DWORD written = 0;
   if ( ! WriteFile ((HANDLE)fd, src, len, &written, NULL)) 
   {
      /*
	WSANOTINITIALISED A successful WSAStartup call must occur
	before using this function.  WSAENETDOWN The network
	subsystem or the associated service provider has failed.
	WSAEAFNOSUPPORT The specified address family is not
	supported.  WSAEINPROGRESS A blocking Windows Sockets 1.1
	call is in progress, or the service provider is still
	processing a callback function.  WSAEMFILE No more socket
	descriptors are available.  WSAENOBUFS No buffer space is
	available. The socket cannot be created.
	WSAEPROTONOSUPPORT The specified protocol is not supported.
	WSAEPROTOTYPE The specified protocol is the wrong type for
	this socket.  WSAESOCKTNOSUPPORT The specified socket type
	is not supported in this address family.
      */
      int errnum=GetLastError();
       file_set_error (err, GetLastError ());
      return -1;
   }
      
   return written;
}

static
gint 
_g_send (gint fd, 
	 gconstpointer src, 
	 size_t len,
	 gint flags, 
	 GError** err )
{
   int res = send (fd, src, len, flags);
   if (res == SOCKET_ERROR)
   {
      /*
	WSANOTINITIALISED A successful WSAStartup call must occur
	before using this function.  WSAENETDOWN The network
	subsystem has failed.  WSAEACCES The requested address is a
	broadcast address, but the appropriate flag was not
	set. Call setsockopt with the SO_BROADCAST parameter to
	allow the use of the broadcast address.  WSAEINTR A
	blocking Windows Sockets 1.1 call was canceled through
	WSACancelBlockingCall.  WSAEINPROGRESS A blocking Windows
	Sockets 1.1 call is in progress, or the service provider is
	still processing a callback function.  WSAEFAULT The buf
	parameter is not completely contained in a valid part of
	the user address space.  WSAENETRESET The connection has
	been broken due to the keep-alive activity detecting a
	failure while the operation was in progress.  WSAENOBUFS No
	buffer space is available.  WSAENOTCONN The socket is not
	connected.  WSAENOTSOCK The descriptor is not a socket.
	WSAEOPNOTSUPP MSG_OOB was specified, but the socket is not
	stream-style such as type SOCK_STREAM, OOB data is not
	supported in the communication domain associated with this
	socket, or the socket is unidirectional and supports only
	receive operations.  WSAESHUTDOWN The socket has been shut
	down; it is not possible to send on a socket after shutdown
	has been invoked with how set to SD_SEND or SD_BOTH.
	WSAEWOULDBLOCK The socket is marked as nonblocking and the
	requested operation would block.  WSAEMSGSIZE The socket is
	message oriented, and the message is larger than the
	maximum supported by the underlying transport.
	WSAEHOSTUNREACH The remote host cannot be reached from this
	host at this time.  WSAEINVAL The socket has not been bound
	with bind, or an unknown flag was specified, or MSG_OOB was
	specified for a socket with SO_OOBINLINE enabled.
	WSAECONNABORTED The virtual circuit was terminated due to a
	time-out or other failure. The application should close the
	socket as it is no longer usable.  WSAECONNRESET The
	virtual circuit was reset by the remote side executing a
	hard or abortive close. For UPD sockets, the remote host
	was unable to deliver a previously sent UDP datagram and
	responded with a "Port Unreachable" ICMP packet. The
	application should close the socket as it is no longer
	usable.  WSAETIMEDOUT The connection has been dropped,
	because of a network failure or because the system on the
	other end went down without notice.
      */
       file_set_error (err, GetLastError ());
      return -1;
   }
      
   return res;
      
}

static
gint 
_g_recv (gint fd,
	 gpointer src, 
	 size_t len,
	 gint flags, 
	 GError** err )
{
   int res = recv (fd, src, len, flags);
   if (res == SOCKET_ERROR)
   {
      /*
	WSANOTINITIALISED A successful WSAStartup call must occur
	before using this function.  WSAENETDOWN The network
	subsystem has failed.  WSAEFAULT The buf parameter is not
	completely contained in a valid part of the user address
	space.  WSAENOTCONN The socket is not connected.  WSAEINTR
	The (blocking) call was canceled through
	WSACancelBlockingCall.  WSAEINPROGRESS A blocking Windows
	Sockets 1.1 call is in progress, or the service provider is
	still processing a callback function.  WSAENETRESET The
	connection has been broken due to the keep-alive activity
	detecting a failure while the operation was in progress.
	WSAENOTSOCK The descriptor is not a socket.  WSAEOPNOTSUPP
	MSG_OOB was specified, but the socket is not stream-style
	such as type SOCK_STREAM, OOB data is not supported in the
	communication domain associated with this socket, or the
	socket is unidirectional and supports only send operations.
	WSAESHUTDOWN The socket has been shut down; it is not
	possible to receive on a socket after shutdown has been
	invoked with how set to SD_RECEIVE or SD_BOTH.
	WSAEWOULDBLOCK The socket is marked as nonblocking and the
	receive operation would block.  WSAEMSGSIZE The message was
	too large to fit into the specified buffer and was
	truncated.  WSAEINVAL The socket has not been bound with
	bind, or an unknown flag was specified, or MSG_OOB was
	specified for a socket with SO_OOBINLINE enabled or (for
	byte stream sockets only) len was zero or negative.
	WSAECONNABORTED The virtual circuit was terminated due to a
	time-out or other failure. The application should close the
	socket as it is no longer usable.  WSAETIMEDOUT The
	connection has been dropped because of a network failure or
	because the peer system failed to respond.  WSAECONNRESET
	The virtual circuit was reset by the remote side executing
	a hard or abortive close. The application should close the
	socket as it is no longer usable. On a UPD-datagram socket
	this error would indicate that a previous send operation
	resulted in an ICMP "Port Unreachable" message.
      */
       file_set_error (err, GetLastError ());
      return -1;
   }
      
   return res;
}

static
gint 
_g_read (gint fd, 
	 gpointer src, 
	 size_t len, 
	 GError** err)
{
   DWORD read = 0;
   if ( ! ReadFile ( (HANDLE) fd, (LPVOID)src, (DWORD)len, &read, NULL))
   {
      /* errors not documented */
       file_set_error (err, GetLastError ());
      return -1;
   }
      
   return read;
}

/* blocking and non-blocking mode */
gint 
g_send (gint fd, 
	gconstpointer src, 
	size_t len,
	gint flags, 
	GError** err )
{
   init_winsock ();

   /* Note: If MSDN documentation would be correct, it should be
      possible to operate with "WriteFile" on sockets, but for some
      reason it failed on my machine (win2000-SP3).  Because of this a
      discriminator is used to distinguish files and sockets. */
   return _g_send (fd, src, len, 0, err);
}

gint 
g_recv (gint fd,
	gpointer src, 
	size_t len,
	gint flags, 
	GError** err )
{
   init_winsock ();

   return _g_recv (fd, src, len, 0, err);
}


/* blocking mode only */
gint 
g_write (gint fd, 
	 gconstpointer src, 
	 size_t len, 
	 GError** err)
{
   init_winsock ();
	
   /* Note: If MSDN documentation would be correct, it should be
      possible to operate with "WriteFile" on sockets, but for some
      reason it failed on my machine (win2000-SP3).  Because of this a
      discriminator is used to distinguish files and sockets. */
   switch (GetFileType ((HANDLE) fd))
   {
   case FILE_TYPE_UNKNOWN:
	    file_set_error (err, EINVAL);
	   return -1;
   case FILE_TYPE_CHAR:
	    file_set_error (err, EINVAL);
	   return -1;
   case FILE_TYPE_DISK:
	   return _g_write (fd, src, len, err);	
   case FILE_TYPE_PIPE: 
	   if (getsockopt (fd, SOL_SOCKET, SO_TYPE, NULL, NULL)!=SO_ERROR)
	   {
			return _g_send (fd, src, len, 0, err);
	   }
	   else
	   {
			return _g_write (fd, src, len, err);	
	   }
   case FILE_TYPE_REMOTE:
	    file_set_error (err, EINVAL);
	   return -1;
   default: 
	    file_set_error (err, EINVAL);
	   return -1;
   }



}



gint 
g_read (gint fd, 
	gpointer src, 
	size_t len, 
	GError** err)
{
   init_winsock ();
 
   /* Note: If MSDN documentation would be correct, it should be
      possible to operate with "WriteFile" on sockets, but for some
      reason it failed on my machine (win2000-SP3).  Because of this a
      discriminator is used to distinguish files and sockets. */
   switch (GetFileType ((HANDLE) fd))
   {
   case FILE_TYPE_UNKNOWN:
	    file_set_error (err, EINVAL);
	   return -1;
   case FILE_TYPE_CHAR:
	    file_set_error (err, EINVAL);
	   return -1;
   case FILE_TYPE_DISK:
	   return _g_read (fd, src, len, err);	
   case FILE_TYPE_PIPE: 
	   if (getsockopt (fd, SOL_SOCKET, SO_TYPE, NULL, NULL)!=SO_ERROR)
	   {
			return _g_recv (fd, src, len, 0, err);
	   }
	   else
	   {
			return _g_read (fd, src, len, err);	
	   }
   case FILE_TYPE_REMOTE:
	    file_set_error (err, EINVAL);
	   return -1;
   default: 
	    file_set_error (err, EINVAL);
	   return -1;
   }
}

gint 
g_ioctl (gint     fd, 
	 gint     cmd,
	 gpointer argp,
	 GError** err)
{
   init_winsock ();

   switch (cmd)
   {
      case FIONBIO:
      {
	 gint res = ioctlsocket ((int) fd, (long) FIONBIO, (u_long*)argp);
	 if (res==SOCKET_ERROR) 
	 {	
	     file_set_error (err, errno);
	    return -1;
	 }
	 return 0;
      }
      default:
	 g_assert ( ! "unknown command");
   }
   return -1;
}
static 
gint 
tcp_pipe (gint fds[2], GError** err)
{
   int res = -1;
   SOCKET insock, outsock, newsock;
   struct sockaddr_in sock_in;
   unsigned int len;

   int fam  = AF_INET;
   int type = SOCK_STREAM;
   int prot = 0;

   /* create the first socket */
   //  newsock = socket (AF_INET, type, 0);
   newsock = socket (fam, type, prot);     
   if (newsock == INVALID_SOCKET)
   {
      goto done;
   }

   /* bind the socket to any unused port */
   sock_in.sin_family      = AF_INET;
   sock_in.sin_port        = 0;
   sock_in.sin_addr.s_addr = htonl (INADDR_LOOPBACK);
   if (bind (newsock, (struct sockaddr *) &sock_in, sizeof (sock_in)) < 0)
   {
      closesocket (newsock);
      goto done;
   }
   len = sizeof (sock_in);
   if (getsockname (newsock, (struct sockaddr *) &sock_in, &len) < 0)
   {
      closesocket (newsock);
      goto done;
   }


   /* For stream sockets, create a listener */
   listen (newsock, 2);

   /* create a connecting socket */
   //  outsock = socket (AF_INET, type, 0);
   outsock = socket (fam, type, 0);
   if (outsock == INVALID_SOCKET)
   {
      closesocket (newsock);
      goto done;
   }

   /* Do a connect */
   if (connect (outsock, (struct sockaddr *) &sock_in,
		sizeof (sock_in)) < 0)
   {
      closesocket (newsock);
      closesocket (outsock);
      goto done;
   }

   /* For stream sockets, accept the connection and close the listener */
   len = sizeof (sock_in);
   insock = accept (newsock, (struct sockaddr *) &sock_in, &len);
   if (insock == INVALID_SOCKET)
   {
      closesocket (newsock);
      closesocket (outsock);
      goto done;
   }
   closesocket (newsock);

   fds[0] = (gint) insock;
   fds[1] = (gint) outsock;

   res = 0; 
   return res;

  done:
    file_set_error (err, WSAGetLastError());
   return res;
}

gint
g_sleep (gint msec, GError** err)
{
   if (msec>=0)
      Sleep (msec);
   else
      Sleep (INFINITE);

   return 0;
}

GQuark
g_process_error_quark (void)
{
   static GQuark quark = 0;
   if (quark == 0)
      quark = g_quark_from_static_string ("g-process-error-quark");
   return quark;
}

static 
void
process_set_error(GError** err, 
		  int      errnum)
{
   gint code = 0;
   if (err==NULL) return;

   /* g_warning ("errcode=%d\n", errnum); */
   switch (errnum)
   {
/*      case EFORK:  code=G_PROCESS_ERROR_FORK; break;  */ /* fork failed due to lack of memory */
/*      case EREAD:  code=G_PROCESS_ERROR_READ; break; */ /* read or select on pipes failed */
/*      case ECHDIR: code=G_PROCESS_ERROR_CHDIR; break; */ /* changing to working dir failed */
      case EACCES: code=G_PROCESS_ERROR_ACCES; break;  /* execv() returned EACCES */
      case EPERM:  code=G_PROCESS_ERROR_PERM; break;   /* execv() returned EPERM */
      case E2BIG:  code=G_PROCESS_ERROR_2BIG; break;   /* execv() returned E2BIG */
      case ENOEXEC:code=G_PROCESS_ERROR_NOEXEC; break; /* execv() returned ENOEXEC */
      case ENAMETOOLONG: code=G_PROCESS_ERROR_NAMETOOLONG; break; /* ""  "" ENAMETOOLONG */
      case ENOENT: code=G_PROCESS_ERROR_NOENT; break;       /* ""  "" ENOENT */
      case ENOMEM: code=G_PROCESS_ERROR_NOMEM; break;       /* ""  "" ENOMEM */
      case ENOTDIR:code=G_PROCESS_ERROR_NOTDIR; break;      /* ""  "" ENOTDIR */
/*      case ELOOP:  code=G_PROCESS_ERROR_LOOP; break;  */      /* ""  "" ELOOP   */
/*      case ETXTBUSY: code=G_PROCESS_ERROR_TXTBUSY; break; */     /* ""  "" ETXTBUSY */
      case EIO:    code=G_PROCESS_ERROR_IO; break;          /* ""  "" EIO */
      case ENFILE: code=G_PROCESS_ERROR_NFILE; break;       /* ""  "" ENFILE */
      case EMFILE: code=G_PROCESS_ERROR_MFILE; break;       /* ""  "" EMFLE */
      case EINVAL: code=G_PROCESS_ERROR_INVAL; break;       /* ""  "" EINVAL */
      case EISDIR: code=G_PROCESS_ERROR_ISDIR; break;       /* ""  "" EISDIR */
/*      case ELIBBAD:code=G_PROCESS_ERROR_LIBBAD; break;  */    /* ""  "" ELIBBAD */
/*      case EFAILED:code=G_PROCESS_ERROR_FAILED; break; */  /* other fatal failure, error->message
							    * should explain
							    */
      default:
	 g_warning ("unknown errcode=%d\n", errnum);
	 code=G_PROCESS_ERROR_FAILED;
   }
   *err = g_error_new_literal (G_PROCESS_ERROR,
			       code,
			       strerror (errnum));
}

gint 
g_process_new (char* const argv[], 
	       gint         fdin,
	       gint         fdout,
	       gint         fderr,
	       guint        flags,
	       GError**     err)
{
   if (init_winsock(err)==-1) return -1;
   {
      STARTUPINFO         siStartInfo;
      PROCESS_INFORMATION piProcInfo;
      
      ZeroMemory (&piProcInfo,  sizeof (PROCESS_INFORMATION));
      ZeroMemory (&siStartInfo, sizeof (STARTUPINFO));
      siStartInfo.cb            = sizeof (STARTUPINFO);
      if (fdin!=-1 || fdout!=-1 || fderr!=-1)
      {
	 siStartInfo.hStdInput     = (HANDLE) ((fdin!=-1)  ? fdin  : G_STDIN);
	 siStartInfo.hStdOutput    = (HANDLE) ((fdout!=-1) ? fdout : G_STDOUT);
	 siStartInfo.hStdError     = (HANDLE) ((fderr!=-1) ? fderr : G_STDERR); 
	 siStartInfo.dwFlags       = STARTF_USESTDHANDLES;
      }

      /* construct command line */
      {
	 /* initial command line, (!must use memory on heap) */
	 GString* cmd_line = g_string_new(argv[0]);

	 int i=0;		  
	 for (i=1; argv[i]; ++i)
	 {
	    cmd_line = g_string_append (cmd_line, " ");     /* seperator */
	    cmd_line = g_string_append (cmd_line, argv[i]);
	 }

	 /* search for executable:             */
	 /* 1.) current working directory.     */
	 /* 2.) system directory               */
	 /* 3.) use PATH environment variable  */
	 if (!CreateProcess (NULL,        // exec name 
			     cmd_line->str,         // command line
			     NULL,                  // process flags
			     NULL,                  // thread flags
			     TRUE,                  // inherit handles
			     0
			     | (flags & G_PROCESS_NEW_GROUP) ? CREATE_NEW_PROCESS_GROUP : 0
			     | (flags & G_PROCESS_DETACH) ? DETACHED_PROCESS : 0
			     ,
			     NULL, 
			     NULL, 
			     &siStartInfo, 
			     &piProcInfo))
	 {
	    g_string_free (cmd_line, TRUE);
	     process_set_error (err, GetLastError ());
	    return -1;
	 }
	 else
	 {
	    g_string_free (cmd_line,TRUE);
	    CloseHandle (piProcInfo.hThread);
	    return (gint) piProcInfo.hProcess; 
	 }
      }
   }
}

/*
  frees resources used to manage process. 
  
  @return -1 on failure, otherwise 0.
*/
gint
g_process_free (gint proc, GError** err)
{
   if ( !CloseHandle ((HANDLE) proc))
   {
      process_set_error (err, GetLastError ());
      return -1;
   }
   return 0;
}

/** @return -1 on error, 0 on timeout, otherwise pid of terminated
    process (equal to pid) 
    @arg status  returns  exit code of process */
gint 
g_process_wait (gint proc, gint* status, gint msecs, GError** err)
{
   switch (WaitForSingleObject (proc, msecs<0 ? INFINITE : msecs))
   {
      case WAIT_ABANDONED:
      case WAIT_OBJECT_0:
      {
	 if ( ! GetExitCodeProcess(obj, status))
	 {
	     process_set_error (err, GetLastError ());
	    ret=-1;
	 }
	 ret=pid;
	 break;
      }
      case WAIT_TIMEOUT:
      default:
	 ret=0;
   }
   return ret;
}

gint 
g_process_kill (gint pid,
		gint sig,
		GError** err)
{
   switch (sig)
   {
/*       case G_SIGINT: */
/* 	 if ( ! GenerateConsoleCtrlEvent (CTRL_C_EVENT, pid)) */
/* 	 { */
/* 	      process_set_error (err, GetLastError ()); */
/* 	    return -1; */
/* 	 } */
/* 	 return 0; */

/*       case G_SIGTERM: */
/* 	 if ( ! GenerateConsoleCtrlEvent (CTRL_BREAK_EVENT, pid)) */
/* 	 { */
/* 	      process_set_error (err, GetLastError ()); */
/* 	    return -1; */
/* 	 } */
/* 	 return 0; */
	 
      case G_SIGKILL:
      {
	 HANDLE obj = OpenProcess (SYNCHRONIZE, FALSE, pid);
	 
	 if (obj==NULL)
	 {
	    /* pid unknown or access denied */
	     process_set_error (err, GetLastError ());
	    return -1;
	 }
	 
	 if ( ! TerminateProcess (obj, 0))
	 {
	     process_set_error (err, GetLastError ());
	    CloseHandle (obj);
	    return -1;
	 }
	 CloseHandle (obj);
	 return 0;
      }
      default:
	 g_assert (! "signal not supported");
		return -1;
   }
	    
}

#endif  /* WIN32 */
